#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2015-2020 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.

"""
Protocol messages between bareos-director and user-agent.
"""

from bareos import __version__
from bareos.bsock.connectiontype import ConnectionType
from bareos.bsock.constants import Constants
from bareos.bsock.protocolmessageids import ProtocolMessageIds
from bareos.bsock.protocolversions import ProtocolVersions


class ProtocolMessages:
    """
    strings defined by the protocol to talk to the Bareos Director.
    """

    def __init__(self, protocolversion=ProtocolVersions.last):
        self.set_version(protocolversion)

    def set_version(self, protocolversion):
        self.protocolversion = protocolversion

    def get_version(self):
        return self.protocolversion

    def hello(self, name, type=ConnectionType.DIRECTOR):
        if type == ConnectionType.FILEDAEMON:
            return bytearray("Hello Director %s calling\n" % (name), "utf-8")
        else:
            if self.protocolversion < ProtocolVersions.bareos_18_2:
                return bytearray("Hello %s calling\n" % (name), "utf-8")
            else:
                return bytearray(
                    "Hello %s calling version %s\n" % (name, __version__), "utf-8"
                )

    # @staticmethod
    # def ok():
    # return "1000 OK:"

    @staticmethod
    def auth_ok():
        return b"1000 OK auth\n"

    @staticmethod
    def auth_failed():
        return b"1999 Authorization failed.\n"

    @staticmethod
    def not_authorized():
        return b"1999 You are not authorized.\n"

    @staticmethod
    def is_auth_ok(msg):
        return msg == ProtocolMessages.auth_ok()

    @staticmethod
    def is_not_authorized(msg):
        return msg == ProtocolMessages.not_authorized()

    @staticmethod
    def pam_user_credentials(pam_username, pam_password):
        """
        Returns a string similar to:
        4002 USERNAME PASSWORD
        """
        return bytearray(
            "{id}{s}{username}{s}{password}".format(
                id=ProtocolMessageIds.PamUserCredentials,
                username=pam_username,
                password=pam_password,
                s=Constants.record_separator,
            ),
            "utf-8",
        )



#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2015-2020 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.

"""Constants used by Bareos."""


class Constants:
    """Constants used by Bareos."""

    unit_separator = chr(0x1F)
    record_separator = chr(0x1E)
    record_separator_compat_regex = r"[ {0}]".format(chr(0x1E))
    group_separator = chr(0x1D)

    # translated enum from
    # https://github.com/bareos/bareos/blob/master/core/src/lib/bsock.h
    BNET_EOD = -1  # ,          /* End of data stream, new data may follow */
    BNET_EOD_POLL = -2  # ,          /* End of data and poll all in one */
    BNET_STATUS = -3  # ,          /* Send full status */
    BNET_TERMINATE = -4  # ,          /* Conversation terminated, doing close() */
    BNET_POLL = -5  # ,          /* Poll request, I'm hanging on a read */
    BNET_HEARTBEAT = -6  # ,          /* Heartbeat Response requested */
    BNET_HB_RESPONSE = -7  # ,          /* Only response permited to HB */
    BNET_xxxxxxPROMPT = -8  # ,          /* No longer used -- Prompt for subcommand */
    BNET_BTIME = -9  # ,          /* Send UTC btime */
    BNET_BREAK = -10  # ,         /* Stop current command -- ctl-c */
    BNET_START_SELECT = -11  # ,         /* Start of a selection list */
    BNET_END_SELECT = -12  # ,         /* End of a select list */
    BNET_INVALID_CMD = -13  # ,         /* Invalid command sent */
    BNET_CMD_FAILED = -14  # ,         /* Command failed */
    BNET_CMD_OK = -15  # ,         /* Command succeeded */
    BNET_CMD_BEGIN = -16  # ,         /* Start command execution */
    BNET_MSGS_PENDING = -17  # ,         /* Messages pending */
    BNET_MAIN_PROMPT = -18  # ,         /* Server ready and waiting */
    BNET_SELECT_INPUT = -19  # ,         /* Return selection input */
    BNET_WARNING_MSG = -20  # ,         /* Warning message */
    BNET_ERROR_MSG = -21  # ,         /* Error message -- command failed */
    BNET_INFO_MSG = -22  # ,         /* Info message -- status line */
    BNET_RUN_CMD = -23  # ,         /* Run command follows */
    BNET_YESNO = -24  # ,         /* Request yes no response */
    BNET_START_RTREE = -25  # ,         /* Start restore tree mode */
    BNET_END_RTREE = -26  # ,         /* End restore tree mode */
    BNET_SUB_PROMPT = -27  # ,         /* Indicate we are at a subprompt */
    BNET_TEXT_INPUT = -28  #          /* Get text input from user */

    _description = {
        BNET_EOD: "End of data stream, new data may follow",
        BNET_EOD_POLL: "End of data and poll all in one",
        BNET_STATUS: "Send full status",
        BNET_TERMINATE: "Conversation terminated",
        BNET_POLL: "Poll request, I'm hanging on a read",
        BNET_HEARTBEAT: "Heartbeat Response requested",
        BNET_HB_RESPONSE: "Only response permited to HB",
        BNET_xxxxxxPROMPT: "No longer used -- Prompt for subcommand",
        BNET_BTIME: "Send UTC btime",
        BNET_BREAK: "Stop current command -- ctl-c",
        BNET_START_SELECT: "Start of a selection list",
        BNET_END_SELECT: "End of a select list",
        BNET_INVALID_CMD: "Invalid command sent",
        BNET_CMD_FAILED: "Command failed",
        BNET_CMD_OK: "Command succeeded",
        BNET_CMD_BEGIN: "Start command execution",
        BNET_MSGS_PENDING: "Messages pending",
        BNET_MAIN_PROMPT: "Server ready and waiting",
        BNET_SELECT_INPUT: "Return selection input",
        BNET_WARNING_MSG: "Warning message",
        BNET_ERROR_MSG: "Error message -- command failed",
        BNET_INFO_MSG: "Info message -- status line",
        BNET_RUN_CMD: "Run command follows",
        BNET_YESNO: "Request yes no response",
        BNET_START_RTREE: "Start restore tree mode",
        BNET_END_RTREE: "End restore tree mode",
        BNET_SUB_PROMPT: "Indicate we are at a subprompt",
        BNET_TEXT_INPUT: "Get text input from user",
    }

    @staticmethod
    def get_description(code):
        """Get a description text for a Bareos constant.

        Args:
           code (int): Bareos ``BNET`` constant.

        Returns:
            str: Description for a Bareos ``BNET`` constant or ``None`` if no description exists.
        """
        try:
            description = Constants._description[code]
        except KeyError:
            # self.logger.error( "unknown bsock code " + code )
            return
        else:
            return description



#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2015-2021 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.

"""
Low Level socket methods to communicate with a Bareos Daemon.
"""

# Authentication code is taken from
# https://github.com/hanxiangduo/bacula-console-python

import hashlib
import hmac
import logging
import random
import re
from select import select
import socket
import ssl
import struct
import sys
import time
import warnings

from bareos.bsock.constants import Constants
from bareos.bsock.connectiontype import ConnectionType
from bareos.bsock.protocolmessageids import ProtocolMessageIds
from bareos.bsock.protocolmessages import ProtocolMessages
from bareos.bsock.protocolversions import ProtocolVersions
from bareos.util.bareosbase64 import BareosBase64
from bareos.util.password import Password
import bareos.exceptions

# Try to load the sslpsk module,
# with implement TLS-PSK (Transport Layer Security - Pre-Shared-Key)
# on top of the ssl module.
# If it is not available, we continue anyway,
# but don't use TLS-PSK.
try:
    import sslpsk
except ImportError:
    warnings.warn(
        u"Connection encryption via TLS-PSK is not available, as the module sslpsk is not installed."
    )


class LowLevel(object):
    """
    Low Level socket methods to communicate with a Bareos Daemon.

    This class should not be used by itself,
    only by inherited classed.
    """

    @staticmethod
    def argparser_get_bareos_parameter(args):
        """Extract arguments.

        This method is usally used together with the method :py:func:`argparser_add_default_command_line_arguments`.

        Args:
           args (ArgParser.Namespace): Arguments retrieved by :py:func:`ArgumentParser.parse_args`.

        Returns:
           dict: The relevant parameter from args to initialize a connection.
        """
        result = {}
        for key, value in vars(args).items():
            if value is not None:
                if key.startswith("BAREOS_"):
                    bareoskey = key.split("BAREOS_", 1)[1]
                    result[bareoskey] = value
        return result

    def __init__(self):
        self.logger = logging.getLogger()
        self.logger.debug("init")
        self.status = None
        self.address = None
        self.password = None
        self.pam_username = None
        self.pam_password = None
        self.port = None
        self.dirname = None
        self.socket = None
        self.auth_credentials_valid = False
        self.max_reconnects = 0
        self.tls_psk_enable = True
        self.tls_psk_require = False
        try:
            self.tls_version = ssl.PROTOCOL_TLS
        except AttributeError:
            self.tls_version = ssl.PROTOCOL_SSLv23
        self.connection_type = None
        self.requested_protocol_version = None
        self.protocol_messages = ProtocolMessages()
        # identity_prefix have to be set in each class
        self.identity_prefix = u"R_NONE"
        self.receive_buffer = b""

    def __del__(self):
        self.close()

    def connect(
        self, address, port, dirname, connection_type, name=None, password=None
    ):
        """Establish a network connection and authenticate.

        Args:
           address (str): Address of the Bareos Director (hostname or IP).

           port (int): Port number of the Bareos Director.

           dirname (str, optional):
              Name of the Bareos Director. Deprecated, normally not required.

           connection_type (int): See :py:class:`bareos.bsock.connectiontype.ConnectionType`.

           name (str, optional):
              Credential name.

           password  (str, bareos.util.Password):
              Credential password, in cleartext or as Password object.

        Returns:
           bool: True, if the authentication succeeds. In earlier versions, authentication failures returned False. However, now an authentication failure raises an exception.

        Raises:
           bareos.exceptions.ConnectionError: If connection can be established.
           bareos.exceptions.PamAuthenticationError: If PAM authentication fails.
           bareos.exceptions.AuthenticationError: If Bareos authentication fails.
        """
        self.address = address
        self.port = int(port)
        if dirname:
            self.dirname = dirname
        else:
            self.dirname = address
        self.connection_type = connection_type
        self.name = name
        if password is None:
            raise bareos.exceptions.ConnectionError(
                u"Parameter 'password' is required."
            )
        if isinstance(password, Password):
            self.password = password
        else:
            self.password = Password(password)

        return self.__connect()

    def __connect(self):
        connected = False
        connected_plain = False
        auth = False
        if self.tls_psk_require:
            if not self.is_tls_psk_available():
                raise bareos.exceptions.ConnectionError(
                    u"TLS-PSK is required, but sslpsk module not loaded/available."
                )
            if not self.tls_psk_enable:
                raise bareos.exceptions.ConnectionError(
                    u"TLS-PSK is required, but not enabled."
                )

        if self.tls_psk_enable and self.is_tls_psk_available():
            try:
                self.__connect_tls_psk()
            except (bareos.exceptions.ConnectionError, ssl.SSLError) as e:
                self._handleSocketError(e)
                if self.tls_psk_require:
                    raise
                else:
                    self.logger.warning(
                        u"Failed to connect via TLS-PSK. Trying plain connection."
                    )
            else:
                connected = True
                self.logger.debug("Encryption: {0}".format(self.socket.cipher()))

        if not connected:
            self.__connect_plain()
            connected = True
            connected_plain = True
            self.logger.debug("Encryption: None")

        if connected:
            try:
                auth = self.auth()
            except bareos.exceptions.PamAuthenticationError:
                raise
            except bareos.exceptions.AuthenticationError:
                if (
                    self.connection_type == ConnectionType.DIRECTOR
                    and self.requested_protocol_version is None
                    and self.get_protocol_version() > ProtocolVersions.bareos_12_4
                ):
                    # reconnect and try old protocol
                    self.logger.warning(
                        "Failed to connect using protocol version {0}. Trying protocol version {1}. ".format(
                            self.get_protocol_version(), ProtocolVersions.bareos_12_4
                        )
                    )
                    self.close()
                    self.__connect_plain()
                    self.protocol_messages.set_version(ProtocolVersions.bareos_12_4)
                    auth = self.auth()
                else:
                    raise

        return auth

    def __connect_plain(self):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # initialize
        try:
            self.socket.connect((self.address, self.port))
        except (socket.error, socket.gaierror) as e:
            self._handleSocketError(e)
            raise bareos.exceptions.ConnectionError(
                "Failed to connect to host {0}, port {1}: {2}".format(
                    self.address, self.port, str(e)
                )
            )

        self.logger.debug("connected to {0}:{1}".format(self.address, self.port))

        return True

    def __connect_tls_psk(self):
        """
        Connect and establish a TLS-PSK connection on top of the connection.
        """
        self.__connect_plain()
        # wrap socket with TLS-PSK
        client_socket = self.socket
        identity = self.get_tls_psk_identity()
        if isinstance(self.password, Password):
            password = self.password.md5()
        else:
            raise bareos.exceptions.ConnectionError(u"No password provided.")
        self.logger.debug("identity = {0}, password = {1}".format(identity, password))
        try:
            self.socket = sslpsk.wrap_socket(
                client_socket,
                ssl_version=self.tls_version,
                ciphers="ALL:!ADH:!LOW:!EXP:!MD5:@STRENGTH",
                psk=(password, identity),
                server_side=False,
            )
        except ssl.SSLError as e:
            # raise ConnectionError(
            #     "failed to connect to host {0}, port {1}: {2}".format(self.address, self.port, str(e)))
            # Using a general raise keep more information about the type of error.
            raise
        return True

    def get_tls_psk_identity(self):
        """Bareos TLS-PSK excepts the identiy is a specific format."""
        name = str(self.name)
        if isinstance(self.name, bytes):
            name = self.name.decode("utf-8")
        result = u"{0}{1}{2}".format(
            self.identity_prefix, Constants.record_separator, name
        )
        return bytes(bytearray(result, "utf-8"))

    @staticmethod
    def is_tls_psk_available():
        """Checks if we have all required modules for TLS-PSK."""
        return "sslpsk" in sys.modules

    def get_protocol_version(self):
        """Get the Bareos Console protocol version that is used.

        Returns:
           int: Number that represents the Bareos Console protocol version (see :py:class:`bareos.bsock.protocolversions.ProtocolVersions`.)
        """
        return self.protocol_messages.get_version()

    def get_cipher(self):
        """
        If a encrypted connection is used, returns information about the encryption. Else it returns None.

        Returns:
           tuple or None: Returns a three-value tuple containing the name of the cipher being used, the version of the SSL protocol that defines its use, and the number of secret bits being used. If the connection is unencrypted or has been established, returns None.
        """
        if hasattr(self.socket, "cipher"):
            return self.socket.cipher()
        else:
            return None

    def auth(self):
        """
        Login to a Bareos Daemon.

        Returns:
           bool: True, if the authentication succeeds. In earlier versions, authentication failures returned False. However, now an authentication failure raises an exception.

        Raises:
           bareos.exceptions.AuthenticationError: if authentication fails.
        """
        bashed_name = self.protocol_messages.hello(self.name, type=self.connection_type)
        # send the bash to the director
        self.send(bashed_name)

        try:
            (ssl, result_compatible, result) = self._cram_md5_respond(
                password=self.password.md5(), tls_remote_need=0
            )
        except bareos.exceptions.SignalReceivedException as e:
            self._handleSocketError(e)
            raise bareos.exceptions.AuthenticationError(
                "Received unexcepted signal: {0}".format(str(e))
            )
        if not result:
            raise bareos.exceptions.AuthenticationError("failed (in response)")
        if not self._cram_md5_challenge(
            clientname=self.name,
            password=self.password.md5(),
            tls_local_need=0,
            compatible=True,
        ):
            raise bareos.exceptions.AuthenticationError("failed (in challenge)")

        self._finalize_authentication()

        return self.auth_credentials_valid

    def receive_and_evaluate_response_message(self):
        """Retrieve a message and evaluate it.

        Only used during in the authentication phase.

        Returns:
           2-tuple: (code, text).
        """
        regex_str = r"^(\d\d\d\d){0}(.*)$".format(
            Constants.record_separator_compat_regex
        )
        regex = bytes(bytearray(regex_str, "utf8"))
        incoming_message = self.recv_msg(regex)
        match = re.search(regex, incoming_message, re.DOTALL)
        code = int(match.group(1))
        text = match.group(2)

        return (code, text)

    def _init_connection(self):
        pass

    def close(self):
        """Close the connection."""
        if self.socket is not None:
            self.socket.close()
        self.socket = None

    def reconnect(self):
        """
        Tries to reconnect.

        Returns:
           bool: True, if the connection could be reestablished.
        """
        result = False
        if self.max_reconnects > 0:
            try:
                self.max_reconnects -= 1
                if self.__connect() and self._init_connection():
                    result = True
            except (socket.error, bareos.exceptions.ConnectionLostError):
                self.logger.warning("failed to reconnect")
        return result

    def call(self, command):
        """Call a Bareos command.

        Args:
           command (str or list): Command to execute. Best provided as a list.

        Returns:
            bytes: Result received from the Daemon.
        """
        if isinstance(command, list):
            command = " ".join(command)
        return self._send_a_command_and_receive_result(command)

    def _send_a_command_and_receive_result(self, command):
        """Send a command and receive the result.

        If connection is lost, try to reconnect.

        Args:
           command (str or list): Command to execute. Best provided as a list.

        Returns:
            bytes: Result received from the Daemon.

        Raises:
            bareos.exceptions.SocketEmptyHeader:
                if an empty header is received.
            bareos.exceptions.ConnectionLostError:
                if the connection is lost-
        """
        result = b""
        try:
            self.send(bytearray(command, "utf-8"))
            result = self.recv_msg()
        except (
            bareos.exceptions.SocketEmptyHeader,
            bareos.exceptions.ConnectionLostError,
        ) as e:
            self.logger.error(
                "connection problem (%s): %s" % (type(e).__name__, str(e))
            )
            if self.reconnect():
                return self._send_a_command_and_receive_result(command, count + 1)
            else:
                raise
        return result

    def send_command(self, command):
        """Alias for :py:func:`call`.

        :deprecated: 15.2.0
        """
        return self.call(command)

    def send(self, msg=None):
        """Send message to the Daemon.

        Args:
           msg (bytearray): Message to send.
        """
        self.__check_socket_connection()
        msg_len = len(msg)  # plus the msglen info

        try:
            # convert to network flow
            self.logger.debug("{0}".format(msg.rstrip()))
            self.socket.sendall(struct.pack("!i", msg_len) + msg)
        except socket.error as e:
            self._handleSocketError(e)

    def recv_bytes(self, length, timeout=10):
        """Receive a number of bytes.

        Args:
          length (int): Number of bytes to receive.
          timeout (float): Timeout in seconds.

        Raises:
            bareos.exceptions.ConnectionLostError:
               If the socket connection gets lost.
            socket.timeout:
               If a timeout occurs on the socket connection,
               meaning no data received.
        """
        self.socket.settimeout(timeout)
        msg = b""
        # get the message
        while length > 0:
            self.logger.debug("expecting {0} bytes.".format(length))
            submsg = self.socket.recv(length)
            if len(submsg) == 0:
                errormsg = u"Failed to retrieve data. Assuming the connection is lost."
                self._handleSocketError(errormsg)
                raise bareos.exceptions.ConnectionLostError(errormsg)
            length -= len(submsg)
            msg += submsg
        return msg

    def recv(self):
        """Receive a single message.

        This is,
        header (4 bytes): if

            * > 0: length of the following message
            * < 0: Bareos signal

        Returns:
           bytearray: Message retrieved via the connection.

        Raises:
            bareos.exceptions.SignalReceivedException:
                If a Bareos signal is received.
        """
        self.__check_socket_connection()
        # get the message header
        header = self.__get_header()
        if header <= 0:
            self.logger.debug("header: " + str(header))
            raise bareos.exceptions.SignalReceivedException(header)
        # get the message
        length = header
        msg = self.recv_submsg(length)
        return msg

    def recv_msg(self, regex=b"^\d\d\d\d OK.*$"):
        """Receive a full message.

        It retrieves messages (header + message text),
        until

           1. the message contains the specified regex or
           2. the header indicates a signal.

        Args:
          regex (bytes): Descripes the expected end of the message.

        Returns:
           bytearray: Message retrieved via the connection.

        Raises:
            bareos.exceptions.SignalReceivedException:
               If a Bareos signal is received.
        """
        self.__check_socket_connection()
        try:
            timeouts = 0
            while True:
                # get the message header
                try:
                    header = self.__get_header()
                except (socket.timeout, ssl.SSLError) as exception:
                    # When using a SSL connection,
                    # a timeout is raised as
                    # ssl.SSLError exception with message: 'The read operation timed out'.
                    # ssl.SSLError is inherited from socket.error.
                    # Because we can't be sure,
                    # that it is really a timeout, we log it.
                    if isinstance(exception, ssl.SSLError) and self.logger.isEnabledFor(
                        logging.DEBUG
                    ):
                        # self.logger.exception('On SSL connections, timeout are raised as ssl.SSLError exceptions:')
                        self.logger.debug("{0}".format(repr(exception)))
                    self.logger.debug("timeout (%i) on receiving header" % (timeouts))
                    timeouts += 1
                else:
                    if header <= 0:
                        # header is a signal
                        self.__set_status(header)
                        if self.is_end_of_message(header):
                            result = self.receive_buffer
                            self.receive_buffer = b""
                            return result
                    else:
                        # header is the length of the next message
                        length = header
                        submsg = self.recv_submsg(length)
                        # check for regex in new submsg
                        # and last line in old message,
                        # which might have been incomplete without new submsg.
                        lastlineindex = self.receive_buffer.rfind(b"\n") + 1
                        self.receive_buffer += submsg
                        match = re.search(
                            regex, self.receive_buffer[lastlineindex:], re.DOTALL
                        )
                        # Bareos indicates end of command result by line starting with 4 digits
                        if match:
                            self.logger.debug(
                                'msg "{0}" matches regex "{1}"'.format(
                                    self.receive_buffer.strip(), regex
                                )
                            )
                            result = self.receive_buffer[
                                0 : lastlineindex + match.end()
                            ]
                            self.receive_buffer = self.receive_buffer[
                                lastlineindex + match.end() + 1 :
                            ]
                            return result
        except socket.error as e:
            self._handleSocketError(e)

    def recv_submsg(self, length):
        """Retrieve a message of the specific length.

        Args:
           length (int): Number of bytes to retrieve.

        Returns:
           bytearray: Retrieved message.
        """
        msg = self.recv_bytes(length)
        if type(msg) is str:
            msg = bytearray(msg.decode("utf-8", 'replace'), "utf-8")
        if type(msg) is bytes:
            msg = bytearray(msg)
        self.logger.debug(str(msg))
        return msg

    def interactive(self):
        """Enter the interactive mode.

        Exit via typing "exit" or "quit".

        Returns:
           bool: True, if exited by user command.
        """
        command = ""
        while command != "exit" and command != "quit" and self.is_connected():
            try:
                command = self._get_input()
            except EOFError:
                return False
            try:
                if command == "exit" or command == "quit":
                    return True
                resultmsg = self.call(command)
                self._show_result(resultmsg)
            except bareos.exceptions.JsonRpcErrorReceivedException as exp:
                print(str(exp))
                # print(str(exp.jsondata))
        return True

    def _get_input(self):
        # Wrapper to retrieve user keyboard input.
        # Python2: raw_input, Python3: input
        try:
            myinput = raw_input
        except NameError:
            myinput = input
        data = myinput(">>")
        return data

    def _show_result(self, msg):
        # print(msg.decode('utf-8'))
        sys.stdout.write(msg.decode("utf-8", 'replace'))
        # add a linefeed, if there isn't one already
        if len(msg) >= 2:
            if msg[-2] != ord(b"\n"):
                sys.stdout.write("\n")

    def __get_header(self, timeout=10):
        header = self.recv_bytes(4, timeout)
        return self.__get_header_data(header)

    def __get_header_data(self, header):
        # struct.unpack:
        #   !: network (big/little endian conversion)
        #   i: integer (4 bytes)
        data = struct.unpack("!i", header)[0]
        return data

    def is_end_of_message(self, data):
        """Checks if a Bareos signal indicates the end of a message.

        Args:
           data (int): Negative integer.

        Returns:
           bool: True, if regular end of message is reached.
        """
        return (
            (not self.is_connected())
            or data == Constants.BNET_EOD
            or data == Constants.BNET_TERMINATE
            or data == Constants.BNET_MAIN_PROMPT
            or data == Constants.BNET_SUB_PROMPT
        )

    def is_connected(self):
        """Verifes that last status still indicates connected.

        Returns:
           bool: True, if still connected.
        """
        return self.status != Constants.BNET_TERMINATE

    def _cram_md5_challenge(
        self, clientname, password, tls_local_need=0, compatible=True
    ):
        """
        client launch the challenge,
        client confirm the dir is the correct director
        """

        # get the timestamp
        # here is the console
        # to confirm the director so can do this on bconsole`way
        rand = random.randint(1000000000, 9999999999)
        # chal = "<%u.%u@%s>" %(rand, int(time.time()), self.dirname)
        chal = "<%u.%u@%s>" % (rand, int(time.time()), clientname)
        msg = bytearray("auth cram-md5 %s ssl=%d\n" % (chal, tls_local_need), "utf-8")
        # send the confirmation
        self.send(msg)
        # get the response
        msg = self.recv()
        if msg[-1] == 0:
            del msg[-1]
        self.logger.debug("received: " + str(msg))

        # hash with password
        hmac_md5 = hmac.new(password, None, hashlib.md5)
        hmac_md5.update(bytes(bytearray(chal, "utf-8")))
        bbase64compatible = BareosBase64().string_to_base64(
            bytearray(hmac_md5.digest()), True
        )
        bbase64notcompatible = BareosBase64().string_to_base64(
            bytearray(hmac_md5.digest()), False
        )
        self.logger.debug("string_to_base64, compatible:     " + str(bbase64compatible))
        self.logger.debug(
            "string_to_base64, not compatible: " + str(bbase64notcompatible)
        )

        is_correct = (msg == bbase64compatible) or (msg == bbase64notcompatible)
        # check against compatible base64 and Bareos specific base64
        if is_correct:
            self.send(ProtocolMessages.auth_ok())
        else:
            self.logger.error(
                "expected result: %s or %s, but get %s"
                % (bbase64compatible, bbase64notcompatible, msg)
            )
            self.send(ProtocolMessages.auth_failed())

        # check the response is equal to base64
        return is_correct

    def _cram_md5_respond(self, password, tls_remote_need=0, compatible=True):
        """
        client connect to dir,
        the dir confirm the password and the config is correct
        """
        # receive from the director
        chal = ""
        ssl = 0
        result = False
        msg = ""
        try:
            msg = self.recv()
        except RuntimeError:
            self.logger.error("RuntimeError exception in recv")
            return (0, True, False)

        # invalid username
        if ProtocolMessages.is_not_authorized(msg):
            self.logger.error("failed: " + str(msg))
            return (0, True, False)

        # check the receive message
        self.logger.debug("(recv): " + str(msg).rstrip())

        msg_list = msg.split(b" ")
        chal = msg_list[2]
        # get th timestamp and the tle info from director response
        ssl = int(msg_list[3][4])
        compatible = True
        # hmac chal and the password
        hmac_md5 = hmac.new((password), None, hashlib.md5)
        hmac_md5.update(bytes(chal))

        # base64 encoding
        msg = BareosBase64().string_to_base64(bytearray(hmac_md5.digest()))

        # send the base64 encoding to director
        self.send(msg)
        received = self.recv()
        if ProtocolMessages.is_auth_ok(received):
            result = True
        else:
            self.logger.error("failed: " + str(received))
        return (ssl, compatible, result)

    def __set_status(self, status):
        self.status = status
        status_text = Constants.get_description(status)
        self.logger.debug(str(status_text) + " (" + str(status) + ")")

    def has_data(self):
        """Is readable data available?

        Returns:
            bool: True: if readable data is available.
        """
        self.__check_socket_connection()
        timeout = 0.1
        readable, writable, exceptional = select([self.socket], [], [], timeout)
        return readable

    def _get_to_prompt(self):
        time.sleep(0.1)
        if self.has_data():
            msg = self.recv_msg()
            self.logger.debug("received message: " + str(msg))
        # TODO: check prompt
        return True

    def __check_socket_connection(self):
        result = True
        if self.socket is None:
            result = False
            if self.auth_credentials_valid:
                # connection have worked before, but now it is gone
                raise bareos.exceptions.ConnectionLostError(
                    "currently no network connection"
                )
            else:
                raise RuntimeError("should connect to director first before send data")
        return result

    def _handleSocketError(self, exception):
        self.logger.warning("socket error: {0}".format(str(exception)))
        self.close()



#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2016-2021 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.

"""
Send and receive the response to Bareos Director Daemon Console interface.
"""

from bareos.bsock.connectiontype import ConnectionType
from bareos.bsock.constants import Constants
from bareos.bsock.lowlevel import LowLevel
from bareos.bsock.protocolmessageids import ProtocolMessageIds
from bareos.bsock.protocolmessages import ProtocolMessages
from bareos.bsock.protocolversions import ProtocolVersions
from bareos.bsock.tlsversionparser import TlsVersionParser
import bareos.exceptions


class DirectorConsole(LowLevel):
    """Send and receive the response to Bareos Director Daemon Console interface.

    Example:
       >>> import bareos.bsock
       >>> directorconsole=bareos.bsock.DirectorConsole(address='localhost', port=9101, name='user1', password='secret')
       >>> print(directorconsole.call('help').decode("utf-8"))
    """

    @staticmethod
    def argparser_add_default_command_line_arguments(argparser):
        """Extend argparser with :py:class:`DirectorConsole` specific parameter.

        Every command line program must offer a similar set of parameter
        to connect to a Bareos Director.
        This method adds the required parameter to an existing ArgParser instance.
        Parameter required to initialize a DirectorConsole class
        are stored in variables prefixed with ``BAREOS_``.

        Use :py:func:`bareos.bsock.lowlevel.LowLevel.argparser_get_bareos_parameter` to retrieve the relevant parameter
        (with the ``BAREOS_`` prefix removed).

        Example:
           >>> argparser = argparse.ArgumentParser(description="Console to Bareos Director.")
           >>> DirectorConsole.argparser_add_default_command_line_arguments(argparser)
           >>> args = argparser.parse_args()
           >>> bareos_args = DirectorConsole.argparser_get_bareos_parameter(args)
           >>> director = DirectorConsole(**bareos_args)

        Args:
          argparser (ArgParser): ArgParser instance.
        """
        argparser.add_argument(
            "--name",
            default="*UserAgent*",
            help='use this to access a specific Bareos director named console. Otherwise it connects to the default console ("*UserAgent*").',
            dest="BAREOS_name",
        )

        argparser.add_argument(
            "-p",
            "--password",
            help="Password to authenticate to a Bareos Director console.",
            required=True,
            dest="BAREOS_password",
        )

        argparser.add_argument(
            "--port",
            default=9101,
            help="Bareos Director network port.",
            dest="BAREOS_port",
        )

        # argparser.add_argument('--dirname', help="Bareos Director name")
        argparser.add_argument(
            "--address",
            default="localhost",
            help="Bareos Director network address.",
            dest="BAREOS_address",
        )

        argparser.add_argument(
            "--protocolversion",
            default=ProtocolVersions.last,
            type=int,
            help=u"Specify the Bareos console protocol version. Default: {0} (current).".format(
                ProtocolVersions.last
            ),
            dest="BAREOS_protocolversion",
        )

        argparser.add_argument(
            "--pam-username",
            help="Username to authenticate against PAM on top off the normal authentication.",
            dest="BAREOS_pam_username",
        )

        argparser.add_argument(
            "--pam-password",
            help="Password to authenticate against PAM on top off the normal authentication.",
            dest="BAREOS_pam_password",
        )

        argparser.add_argument(
            "--tls-psk-require",
            help="Allow only encrypted connections. Default: False.",
            action="store_true",
            dest="BAREOS_tls_psk_require",
        )

        TlsVersionParser().add_argument(argparser)

    def __init__(
        self,
        address="localhost",
        port=9101,
        dirname=None,
        name="*UserAgent*",
        password=None,
        protocolversion=None,
        pam_username=None,
        pam_password=None,
        tls_psk_enable=True,
        tls_psk_require=False,
        tls_version=None,
    ):
        """\

        Args:
           address (str): Address of the Bareos Director (hostname or IP).

           port (int): Port number of the Bareos Director.

           dirname (str, optional):
              Name of the Bareos Director. Deprecated, normally not required.

           name (str, optional):
              Name of the Director Console. Leave empty when connecting to the Bareos Default Console.

           password  (str, bareos.util.Password):
              Password, in cleartext or as Password object.

           protocolversion (None, bareos.bsock.ProtocolVersions.last, bareos.bsock.ProtocolVersions.bareos_12_4, bareos.bsock.ProtocolVersions.bareos_18_2):
              Specify the Bareos Console protocol version to use.

           pam_username (str): Additional username when using PAM.

           pam_password (str): Additional user password when using PAM.

           tls_psk_enable (boolean): Enable TLS-PSK.

           tls_psk_require (boolean): Enforce using TLS-PSK.

           tls_version (None, ssl.PROTOCOL_TLS, ssl.PROTOCOL_TLSv1, ssl.PROTOCOL_TLSv1_1, ssl.PROTOCOL_TLSv1_2):
              TLS protocol version to use.

        Raises:
          bareos.exceptions.ConnectionError: On connections errors.
        """

        super(DirectorConsole, self).__init__()
        self.pam_username = pam_username
        self.pam_password = pam_password
        self.tls_psk_enable = tls_psk_enable
        self.tls_psk_require = tls_psk_require
        if tls_version is not None:
            self.tls_version = tls_version
        self.identity_prefix = u"R_CONSOLE"
        if protocolversion is not None:
            self.requested_protocol_version = int(protocolversion)
            self.protocol_messages.set_version(self.requested_protocol_version)
        self.connect(address, port, dirname, ConnectionType.DIRECTOR, name, password)
        self._init_connection()
        self.max_reconnects = 1

    def _finalize_authentication(self):
        code, text = self.receive_and_evaluate_response_message()

        self.logger.debug(u"code: {0}".format(code))

        #
        # Test if PAM is requested.
        # If yes, handle PAM messages.
        #
        # Is is only available, with protocolversion >= ProtocolVersions.bareos_18_2,
        # however as it is optional,
        # it will be evaluated with all protocol versions.
        #
        if code == ProtocolMessageIds.PamRequired:
            self.logger.debug(u"PAM request: {0}".format(text))
            if (not self.pam_username) or (not self.pam_password):
                raise bareos.exceptions.PamAuthenticationError(
                    "PAM authentication is requested, but no PAM credentials given. Giving up.\n"
                )
            self.send(
                ProtocolMessages.pam_user_credentials(
                    self.pam_username, self.pam_password
                )
            )
            try:
                code, text = self.receive_and_evaluate_response_message()
            except bareos.exceptions.ConnectionLostError as e:
                raise bareos.exceptions.PamAuthenticationError(
                    u"PAM authentication failed."
                )
        else:
            if (self.pam_username) or (self.pam_password):
                raise bareos.exceptions.PamAuthenticationError(
                    "PAM credentials provided, but this Director console does not offer PAM login. Giving up.\n"
                )

        #
        # Test if authentication has been accepted.
        #
        if code == ProtocolMessageIds.Ok:
            self.logger.info(u"Authentication: {0}".format(text))
            self.auth_credentials_valid = True
        else:
            raise bareos.exceptions.AuthenticationError(
                "Received unexcepted message: {0} {1} (expecting auth ok)".format(
                    code, text
                )
            )

        if self.get_protocol_version() >= ProtocolVersions.bareos_18_2:
            #
            # Handle info message.
            #
            code, text = self.receive_and_evaluate_response_message()
            if code == ProtocolMessageIds.InfoMessage:
                self.logger.debug(u"Info: {0}".format(text))
            else:
                raise bareos.exceptions.AuthenticationError(
                    "Received unexcepted message: {0} {1} (expecting info message)".format(
                        code, text
                    )
                )

    def _init_connection(self):
        self.call("autodisplay off")

    def _get_to_prompt(self):
        self.send(b".")
        return super(DirectorConsole, self)._get_to_prompt()



#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2019-2021 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.

"""
Bareos Protocol Versions.
"""


class ProtocolVersions(object):
    """
    Bareos Protocol Versions.
    """

    bareos_12_4 = 1
    bareos_18_2 = 2
    last = 2



#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2015-2021 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.

"""Deprecated, use :py:class:`bareos.bsock.DirectorConsole` instead."""

from bareos.bsock.directorconsole import DirectorConsole


class BSock(DirectorConsole):
    """Communicates with the bareos-director.

    :deprecated: 15.2.0

    Deprecated, use :py:class:`bareos.bsock.DirectorConsole` instead.
    """

    def __init__(self, *args, **kwargs):
        super(BSock, self).__init__(*args, **kwargs)



#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2015-2021 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.

"""
Bareos specific base64 implementation.

Bacula and therefore Bareos specific implementation of a base64 decoder.
"""


class BareosBase64(object):
    """Bareos specific base64 implementation."""

    base64_digits = list(
        "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"
    )

    def __init__(self):
        """Initialize the Base 64 conversion routines."""
        self.base64_map = dict(list(zip(self.base64_digits, list(range(0, 64)))))

    @staticmethod
    def twos_comp(val, bits):
        """Compute the 2's compliment of int value val."""
        if (val & (1 << (bits - 1))) != 0:
            val = val - (1 << bits)
        return val

    def base64_to_int(self, base64):
        """Convert a base 64 string to integer.

        Args:
           base64 (str): base 64 string.

        Returns:
           int: Integer value of the base64 string.
        """
        value = 0
        first = 0
        neg = False

        if base64[0] == "-":
            neg = True
            first = 1

        for i in range(first, len(base64)):
            value = value << 6
            try:
                value += self.base64_map[base64[i]]
            except KeyError:
                print("KeyError:", i)

        return -value if neg else value

    def int_to_base64(self, value):
        """Convert an integer to base 64.

        Args:
           value (int): integeer value.

        Returns:
           str: base 64 representation of value.
        """
        result = ""
        if value < 0:
            result = "-"
            value = -value

        while value:
            charnumber = value % 0x3F
            result += self.base64_digits[charnumber]
            value = value >> 6
        return result

    def string_to_base64(self, string, compatible=False):
        """Convert a string to base64.

        Args:
           string (str): string to be converted.
           compatible (bool): If True, generate Baculas broken version of base 64 strings.

        Returns:
           bytearray: base 64 representation of the given string.
        """
        buf = ""
        reg = 0
        rem = 0
        char = 0
        i = 0
        while i < len(string):
            if rem < 6:
                reg <<= 8
                char = string[i]
                if not compatible:
                    if char >= 128:
                        char = self.twos_comp(char, 8)
                reg |= char
                i += 1
                rem += 8

            save = reg
            reg >>= rem - 6
            buf += self.base64_digits[reg & 0x3F]
            reg = save
            rem -= 6

        if rem:
            mask = (1 << rem) - 1
            if compatible:
                buf += self.base64_digits[(reg & mask) << (6 - rem)]
            else:
                buf += self.base64_digits[reg & mask]
        return bytearray(buf, "utf-8")



#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2015-2021 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.

"""
Handle file paths.
"""

from copy import copy


class Path(object):
    """
    Class to handle file paths.
    """

    def __init__(self, path=None):
        """\

        Args:
           path (str, optional): string representation of the file system path.
        """
        self.__set_defaults()
        self.set_path(path)

    def __str__(self):
        result = ""
        if self.is_root():
            result += "/"
        result += "/".join(self.path)
        if (not self.is_root()) or self.len() > 0:
            if self.is_directory():
                result += "/"
        return result

    def __set_defaults(self):
        self.path_orig = None
        self.root = False
        self.directory = False
        self.path = None

    def set_path(self, path):
        if path == None:
            self.__set_defaults()
        elif isinstance(path, str):
            self.path_orig = path
            components = self.path_orig.split("/")
            self.path = [i for i in components if i != ""]
            if path == "":
                self.root = False
                self.directory = True
            else:
                self.root = False
                if self.path_orig[0] == "/":
                    self.root = True
                self.directory = False
                if components[-1] == "":
                    self.directory = True
        else:
            # exception
            pass

    def get(self, index=None):
        if index == None:
            return self.path
        else:
            return self.path[index]

    def shift(self):
        """
        Removes the first component of the path.

        Example:

           .. code:: python

              >>> path = Path("/usr/bin/python")
              >>> path.shift()
              'usr'
              >>> print(path)
              /bin/python

        Returns:
            str: First component of the path.

        Raises:
            IndexError: if path can't be shifted (path is empty).
        """
        result = self.get(0)
        self.remove(0)
        return result

    def is_directory(self):
        return self.directory

    def is_root(self):
        return self.root

    def remove(self, index):
        del self.path[index]

    def len(self):
        return len(self.path)



#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2020-2021 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.

"""
Classes to parse TLS version paramters.
"""

import argparse
from collections import OrderedDict
import ssl


class ArgParserTlsVersionAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, getattr(ssl, self.choices.get(values)))


class TlsVersionParser:
    def __init__(self):
        # Add the possibility to specify the TLS protocol version.
        # This is required,
        # as sslpsk (1.0.0), depending on the Python and openssl version
        # is known to fail on various protocol versions,
        # especially with the default (PROTOCOL_TLS).
        # Anyhow, if possible, use the default (PROTOCOL_TLS),
        # as this covers different protocol versions,
        # including all versions >= v1.3.
        # There will be no specific constant TLS >= 1.3.
        self.tls_version_options = {
            # "default": "PROTOCOL_TLS",
            "v1": "PROTOCOL_TLSv1",
            "v1.1": "PROTOCOL_TLSv1_1",
            "v1.2": "PROTOCOL_TLSv1_2",
        }

        # remove invalid options
        for key, value in self.tls_version_options.items():
            if not hasattr(ssl, value):
                del self.tls_version_options[key]

    def add_argument(self, argparser):
        argparser.add_argument(
            "--tls-version",
            help="Use a specific TLS protocol version.",
            action=ArgParserTlsVersionAction,
            choices=OrderedDict(sorted(self.tls_version_options.items())),
            dest="BAREOS_tls_version",
        )

    def get_protocol_version_from_string(self, tls_version):
        if tls_version is None:
            return None
        result = None
        try:
            result = getattr(ssl, self.tls_version_options[tls_version.lower()])
        except (KeyError, AttributeError) as exc:
            pass
        return result

    def get_protocol_versions(self):
        return self.tls_version_options.keys()



#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2016-2020 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.

"""Module to access a http://www.bareos.org backup system.

.. note::

   By default, the Bareos Director (>= 18.2.4) uses TLS-PSK when communicating through the network.

   Unfortunately the Python core module ``ssl``
   does not support TLS-PSK.
   For testing this module should be used without TLS.
   The section `Transport Encryption (TLS-PSK)`_ describes
   how to use TLS-PSK and about the limitations.


Preparations
============

Create some named consoles for testing:

.. code-block:: shell-session

   root@host:~# bconsole
   *configure add console name=user1 password=secret profile=operator TlsEnable=no
   *configure add console name=user-tls password=secret profile=operator


This creates a console user with name `user1` and the profile `operator`.
The `operator` profile is a default profile that comes with the Bareos Director.
It does allow most commands. It only deny some dangerous commands (see ``show profile=operator``),
so it is well suited for this purpose.
Futhermore, TLS enforcement is disabled for this console user.

For testing with TLS-PSK, we also create the user `user-tls`.


Examples
========

Calling bareos-director console commands
----------------------------------------

.. code:: python

   >>> import bareos.bsock
   >>> directorconsole=bareos.bsock.DirectorConsole(address='localhost', port=9101, name='user1', password='secret')
   >>> print(directorconsole.call('help').decode("utf-8"))

This creates a console connection to a Bareos Director.
This connection can be used to `call` commands.
These are the same commands as available via ``bconsole``.

To connect to the default console instead, omit the `name` parameter:

.. code:: python

   >>> directorconsole = bareos.bsock.DirectorConsole(address='localhost', port=9101, password='defaultconsolepassword')

The result of the call method is a ``bytes`` object. In most cases, it has to be decoded to UTF-8.



Simple version of the bconsole in Python
----------------------------------------

.. code:: python

   >>> import bareos.bsock
   >>> directorconsole = bareos.bsock.DirectorConsole(address='localhost', port=9101, password='secret')
   >>> directorconsole.interactive()

Or use the ``bconsole.py`` script:

.. code-block:: shell-session

   bconsole.py --debug --name=user1 --password=secret localhost


Use JSON objects of the API mode 2
----------------------------------

Requires: Bareos >= 15.2

The class `DirectorConsoleJson` is inherited from `DirectorConsole`
and uses the Director Console API mode 2 (JSON).

For general information about API mode 2 and what data structures to expect,
see https://docs.bareos.org/DeveloperGuide/api.html#api-mode-2-json

Example:

.. code:: python

   >>> import bareos.bsock
   >>> directorconsole = bareos.bsock.DirectorConsoleJson(address='localhost', port=9101, password='secret')
   >>> pools = directorconsole.call('list pools')
   >>> for pool in pools["pools"]:
   ...   print(pool["name"])
   ...
   Scratch
   Incremental
   Full
   Differential

The results the the `call` method is a ``dict`` object.

In case of an error, an exception, derived from :py:class:`bareos.exceptions.Error` is raised.

Example:


.. code:: python

   >>> directorconsole.call("test it")
   Traceback (most recent call last):
   ...
   bareos.exceptions.JsonRpcErrorReceivedException: failed: test it: is an invalid command.



.. _section-python-bareos-tls-psk:

Transport Encryption (TLS-PSK)
==============================

Since Bareos >= 18.2.4, Bareos supports TLS-PSK (Transport-Layer-Security Pre-Shared-Key) to secure its network connections and uses this by default.

Unfortenatly, the Python core module `ssl` does not support TLS-PSK.
There is limited support by the extra module `sslpsk` (see https://github.com/drbild/sslpsk).

Fallback To Unencrypted Connections
-----------------------------------

In order to work in most cases, even if ``sslpsk`` is not available,
the `DirectorConsole` uses a fallback.
If connecting via TLS-PSK fails, it falls back to the old, unencrypted protocol version.
In this case, a warning is issued, but the connection will work nevertheless:

.. code:: python

   >>> import bareos.bsock
   /.../bareos/bsock/lowlevel.py:39: UserWarning: Connection encryption via TLS-PSK is not available, as the module sslpsk is not installed.
   >>> directorconsole=bareos.bsock.DirectorConsole(address='localhost', port=9101, name='user-tls', password='secret')
   socket error: Conversation terminated (-4)
   Failed to connect using protocol version 2. Trying protocol version 1.
   >>> print(directorconsole.call('help').decode("utf-8"))


To enforce a encrypted connection, use the ``tls_psk_require=True`` parameter:

.. code:: python

   >>> import bareos.bsock
   /.../bareos/bsock/lowlevel.py:39: UserWarning: Connection encryption via TLS-PSK is not available, as the module sslpsk is not installed.
   >>> directorconsole=bareos.bsock.DirectorConsole(address='localhost', port=9101, name='user-tls', password='secret', tls_psk_require=True)
   Traceback (most recent call last):
   ...
   bareos.exceptions.ConnectionError: TLS-PSK is required, but sslpsk module not loaded/available.


In this case, an exception is raised, if the connection can not be established via TLS-PSK.

sslpsk
------

The extra module `sslpsk` (see https://github.com/drbild/sslpsk)
extends the core module `ssl` by TLS-PSK.

At the time of writing, the lasted version installable via pip is 1.0.0 (https://pypi.org/project/sslpsk/), which is not working with Python >= 3.

If `python-bareos` should use TLS-PSK with Python >= 3,
the latest version must by installed manually:

.. code:: shell

   git clone https://github.com/drbild/sslpsk.git
   cd sslpsk
   python setup.py build
   python setup.py install

`python-bareos` will detect, that `sslpsk` is available and will use it automatically.
This can be verified by following command:

.. code:: python

   >>> import bareos.bsock
   >>> bareos.bsock.DirectorConsole.is_tls_psk_available()
   True

Another limitation of the current `sslpsk` version is,
that it is not able to autodetect the TLS protocol version to use.

In order to use it, specify ``tls_version`` with an appropriate protocol version.
In most cases this should be ``tls_version=ssl.PROTOCOL_TLSv1_2``,
like in the following example:

.. code:: python

   >>> import ssl
   >>> import bareos.bsock
   >>> directorconsole = bareos.bsock.DirectorConsoleJson(address='localhost', user='user-tls', password='secret', tls_version=ssl.PROTOCOL_TLSv1_2)
   >>> print(directorconsole.call('help').decode("utf-8"))
"""

from bareos.exceptions import *
from bareos.util.password import Password
from bareos.bsock.connectiontype import ConnectionType
from bareos.bsock.constants import Constants
from bareos.bsock.filedaemon import FileDaemon
from bareos.bsock.directorconsole import DirectorConsole
from bareos.bsock.directorconsolejson import DirectorConsoleJson
from bareos.bsock.protocolversions import ProtocolVersions
from bareos.bsock.tlsversionparser import TlsVersionParser

# compat
from bareos.bsock.bsock import BSock
from bareos.bsock.bsockjson import BSockJson



#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2015-2021 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.

"""
Bareos specific exceptions.
"""

from bareos.bsock.constants import Constants


class Error(Exception):
    """Base Bareos error exception."""

    pass


class ConnectionError(Error):
    """Error with the connection."""

    pass


class ConnectionLostError(Error):
    """Connection lost error."""

    pass


class SocketEmptyHeader(Error):
    """Socket connection received an empty header. Connection lost?"""

    pass


class AuthenticationError(ConnectionError):
    """Error during authentication."""

    pass


class PamAuthenticationError(AuthenticationError):
    """Error during PAM authentication."""

    pass


class SignalReceivedException(Error):
    """Received a Bareos signal during a connection."""

    def __init__(self, signal):
        # Call the base class constructor with the parameters it needs
        message = Constants.get_description(signal)
        super(SignalReceivedException, self).__init__(
            "{0} ({1})".format(message, signal)
        )

        # Now for your custom code...
        self.signal = signal


class JsonRpcErrorReceivedException(Error):
    """Received JSON-RPC error object.

    This exception is raised,
    if a JSON-RPC error object is received.
    """

    def __init__(self, jsondata):
        # Call the base class constructor with the parameters it needs

        # Expected result will look like this:
        #
        # {u'jsonrpc': u'2.0', u'id': None, u'error': {u'message': u'failed', u'code': 1, u'data': {u'messages': {u'error': [u'INVALIDCOMMAND: is an invalid command.\n']}, u'result': {}}}}

        try:
            message = jsondata["error"]["message"]
        except KeyError:
            message = ""
        try:
            errormessages = jsondata["error"]["data"]["messages"]["error"]
            error = "".join(errormessages)
        except (KeyError, TypeError):
            error = str(jsondata)

        super(JsonRpcErrorReceivedException, self).__init__(
            "{0}: {1}".format(message, error)
        )

        # Now for your custom code...
        self.jsondata = jsondata


class JsonRpcInvalidJsonReceivedException(JsonRpcErrorReceivedException):
    """Received invalid JSON-RPC object.

    This exception is raised,
    if a invalid JSON-RPC object is received (e.g. data is not a valid JSON structure).
    """

    def __init__(self, jsondata):
        # Call the base class constructor with the parameters it needs

        # Expected result will look like this:
        #
        # {'error': {'message': 'No JSON object could be decoded', 'code': 2, 'data': bytearray(b'Client {\n  Name = "bareos-fd"\n  Description = "Client resource of the Director itself."\n  Address = "localhost"\n  Password = "****************"\n}\n\n{"jsonrpc":"2.0","id":null,"result":{}}')}}

        try:
            message = jsondata["error"]["message"]
        except KeyError:
            message = ""
        try:
            origdata = str(jsondata["error"]["data"])
        except (KeyError, TypeError):
            origdata = ""

        super(JsonRpcErrorReceivedException, self).__init__(
            "{0}: {1}".format(message, origdata)
        )

        # Now for your custom code...
        self.jsondata = jsondata



#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2015-2021 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.

"""
Bareos utility classes.
"""

from bareos.util.bareosbase64 import BareosBase64
from bareos.util.password import Password
from bareos.util.path import Path

__all__ = ["BareosBase64", "Password", "Path"]



#!/usr/bin/python
#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2019-2020 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.


import os
import re
from setuptools import find_packages, setup


def get_version():
    base_dir = os.path.abspath(os.path.dirname(__file__))

    try:
        with open(os.path.join(base_dir, "bareos", "VERSION.txt")) as version_file:
            # read version
            # and adapt it according to
            # https://www.python.org/dev/peps/pep-0440/.
            fullversion = version_file.read().strip()
            __version__ = re.compile(r"~pre([0-9]+).*").sub(r".dev\1", fullversion)
    except IOError:
        # Fallback version.
        # First protocol implemented
        # has been introduced with this version.
        __version__ = "18.2.5"

    return __version__


setup(
    name="python-bareos",
    version=get_version(),
    license="AGPLv3",
    author="Bareos Team",
    author_email="packager@bareos.com",
    packages=find_packages(),
    scripts=["bin/bconsole.py", "bin/bconsole-json.py", "bin/bareos-fd-connect.py"],
    package_data={"bareos": ["VERSION.txt"]},
    url="https://github.com/bareos/bareos/",
    # What does your project relate to?
    keywords="bareos",
    description="Client library and tools for Bareos console access.",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    # Python 2.6 is used by RHEL/Centos 6.
    # When RHEL/Centos 6 is no longer supported (End of 2020),
    # Python 2.6 will no longer be supported by python-bareos.
    python_requires=">=2.6",
    extras_require={"TLS-PSK": ["sslpsk"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: System :: Archiving :: Backup",
    ],
)



#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2016-2021 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.

"""
Send and receive the response to Bareos File Daemon (bareos-fd).
"""

from bareos.bsock.connectiontype import ConnectionType
from bareos.bsock.lowlevel import LowLevel
from bareos.bsock.protocolmessageids import ProtocolMessageIds
from bareos.bsock.tlsversionparser import TlsVersionParser
import bareos.exceptions
import shlex


class FileDaemon(LowLevel):
    """Send and receive the response to Bareos File Daemon (bareos-fd)."""

    @staticmethod
    def argparser_add_default_command_line_arguments(argparser):
        """Extend argparser with :py:class:`FileDaemon` specific parameter.

        Every command line program must offer a similar set of parameter
        to connect to a Bareos File Daemon.
        This method adds the required parameter to an existing ArgParser instance.
        Parameter required to initialize a FileDaemon class
        are stored in variables prefixed with ``BAREOS_``.

        Use :py:func:`bareos.bsock.lowlevel.LowLevel.argparser_get_bareos_parameter` to retrieve the relevant parameter
        (with the ``BAREOS_`` prefix removed).

        Example:

           .. code-block:: python

              argparser = argparse.ArgumentParser(description='Connect to  Bareos File Daemon.')
              FileDaemon.argparser_add_default_command_line_arguments(argparser)
              args = argparser.parse_args()
              bareos_args = DirectorConsole.argparser_get_bareos_parameter(args)
              fd = FileDaemon(**bareos_args)

        Args:
          argparser (ArgParser): ArgParser instance.
        """
        argparser.add_argument(
            "--name",
            help="Name of the Director resource in the File Daemon.",
            required=True,
            dest="BAREOS_name",
        )

        argparser.add_argument(
            "-p",
            "--password",
            help="Password to authenticate to a Bareos File Daemon.",
            required=True,
            dest="BAREOS_password",
        )

        argparser.add_argument(
            "--port",
            default=9102,
            help="Bareos File Daemon network port.",
            dest="BAREOS_port",
        )

        argparser.add_argument(
            "--address",
            default="localhost",
            help="Bareos File Daemon network address.",
            dest="BAREOS_address",
        )

        argparser.add_argument(
            "--tls-psk-require",
            help="Allow only encrypted connections. Default: False.",
            action="store_true",
            dest="BAREOS_tls_psk_require",
        )

        TlsVersionParser().add_argument(argparser)

    def __init__(
        self,
        address="localhost",
        port=9102,
        name=None,
        password=None,
        tls_psk_enable=True,
        tls_psk_require=False,
        tls_version=None,
    ):
        """\

        Args:
           address (str): Address of the Bareos File Daemon (hostname or IP).

           port (int): Port number of the Bareos File Daemon.

           name (str):
              Name of the File Daemon.

           password  (str, bareos.util.Password):
              Password, in cleartext or as Password object.

           tls_psk_enable (boolean): Enable TLS-PSK.

           tls_psk_require (boolean): Enforce using TLS-PSK.

           tls_version (None, ssl.PROTOCOL_TLS, ssl.PROTOCOL_TLSv1, ssl.PROTOCOL_TLSv1_1, ssl.PROTOCOL_TLSv1_2):
              TLS protocol version to use.

        Raises:
          bareos.exceptions.ConnectionError: On connections errors.
        """
        super(FileDaemon, self).__init__()
        self.tls_psk_enable = tls_psk_enable
        self.tls_psk_require = tls_psk_require
        if tls_version is not None:
            self.tls_version = tls_version
        # Well, we are not really a Director,
        # but using the interface provided for Directors.
        self.identity_prefix = u"R_DIRECTOR"
        self.connect(address, port, dirname, ConnectionType.FILEDAEMON, name, password)
        self._init_connection()

    def _finalize_authentication(self):
        code, text = self.receive_and_evaluate_response_message()

        self.logger.debug(u"code: {0}".format(code))

        #
        # Test if authentication has been accepted.
        #
        if code == ProtocolMessageIds.FdOk:
            self.logger.info(u"Authentication: {0}".format(text))
            self.auth_credentials_valid = True
        else:
            raise bareos.exceptions.AuthenticationError(
                "Received unexcepted message: {0} {1} (expecting auth ok)".format(
                    code, text
                )
            )

    def call(self, command):
        """Calls a command on the Bareos File Daemon and returns its result.

        Args:
           command (str or list): Command to execute. Best provided as a list.

        Returns:
            bytes: Result received from the File Daemon.
        """
        if isinstance(command, list):
            cmdlist = command
        else:
            cmdlist = shlex.split(command)
        command0 = []
        for arg in cmdlist:
            command0.append(arg.replace(" ", "\x01"))
        return super(FileDaemon, self).call(command0)



#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2015-2021 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.

"""
Handle Bareos passwords.
"""

import hashlib


class Password(object):
    """
    Handle Bareos passwords.

    Instead of working with plain text passwords,
    Bareos often uses the MD5 hash of the password.

    This class make a password available in both forms.
    """

    def __init__(self, password=None):
        """\

        Args:
           password (str, optional): Password as plain text.
        """
        self.password_md5 = None
        self.set_plaintext(password)

    def set_plaintext(self, password):
        """Set new password.

        Args:
           password (str): Password as plain text.
        """
        self.password_plaintext = bytearray(password, "utf-8")
        self.set_md5(self.__plaintext2md5(password))

    def set_md5(self, password):
        """Set the MD5 form of the password.

        If the password is initialized by this method,
        it is only available as MD5 hash, not as plain text password.

        Args:
           password (bytes): Password as MD5 hash.
        """
        self.password_md5 = password

    def plaintext(self):
        """Get password as plain text.

        Returns:
            str: Password as plain text.
        """
        return self.password_plaintext.decode("utf-8")

    def md5(self):
        """Get password as MD5 hash.

        Returns:
            bytes: Password as MD5 hash.
        """
        return self.password_md5

    @staticmethod
    def __plaintext2md5(password):
        """
        md5 the password and return the hex style
        """
        md5 = hashlib.md5()
        md5.update(bytes(bytearray(password, "utf-8")))
        return bytes(bytearray(md5.hexdigest(), "utf-8"))



#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2016-2021 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.

"""
Communicate with the Bareos Director Daemon Console interface in API mode 2 (JSON).
"""

from bareos.bsock.directorconsole import DirectorConsole
import bareos.exceptions
from pprint import pformat, pprint
import json


class DirectorConsoleJson(DirectorConsole):
    """Communicate with the Bareos Director Daemon Console interface in API mode 2 (JSON).

    Example:

       >>> import bareos.bsock
       >>> directorconsole = bareos.bsock.DirectorConsoleJson(address='localhost', port=9101, password='secret')
       >>> pools = directorconsole.call('list pools')
       >>> for pool in pools["pools"]:
       ...   print(pool["name"])
       ...
       Scratch
       Incremental
       Full
       Differential

       The results the the `call` method is a ``dict`` object.

       In case of an error, an exception, derived from ``bareos.exceptions.Error`` is raised.
    """

    def __init__(self, *args, **kwargs):
        """\
        **Parameters:** The parameter are identical to :py:class:`bareos.bsock.directorconsole.DirectorConsole`.

        Raises:
            bareos.exceptions.JsonRpcInvalidJsonReceivedException:
                if the ".api" command is not available.
        """
        super(DirectorConsoleJson, self).__init__(*args, **kwargs)

    def _init_connection(self):
        # older version did not support compact mode,
        # therfore first set api mode to json (which should always work in bareos >= 15.2.0)
        # and then set api mode json compact (which should work with bareos >= 15.2.2)
        self.logger.debug(self.call(".api json"))
        self.logger.debug(self.call(".api json compact=yes"))

    def call(self, command):
        """Calls a command on the Bareos Director and returns its result.

        If the JSON-RPC result indicates an error
        (contains the ``error`` element),
        an exception will be raised.

        Args:
           command (str or list): Command to execute. Best provided as a list.

        Returns:
            dict: Result received from the Bareos Director.

        Raises:
            bareos.exceptions.JsonRpcErrorReceivedException:
                if an JSON-RPC error object is received.
            bareos.exceptions.JsonRpcInvalidJsonReceivedException:
                if an invalid JSON-RPC result is received.
        """
        json = self.call_fullresult(command)
        if json == None:
            return
        if "result" in json:
            result = json["result"]
        elif "error" in json:
            raise bareos.exceptions.JsonRpcErrorReceivedException(json)
        else:
            raise bareos.exceptions.JsonRpcInvalidJsonReceivedException(json)
        return result

    def call_fullresult(self, command):
        """Calls a command on the Bareos Director and returns its result.

        Returns:
            dict: Result received from the Bareos Director.

        Raises:
            bareos.exceptions.JsonRpcInvalidJsonReceivedException:
                if an invalid JSON-RPC result is received.
        """
        resultstring = super(DirectorConsoleJson, self).call(command)
        data = None
        if resultstring:
            try:
                data = json.loads(resultstring.decode("utf-8"))
            except ValueError as e:
                # in case result is not valid json,
                # create a JSON-RPC wrapper
                data = {"error": {"code": 2, "message": str(e), "data": resultstring}}
                raise bareos.exceptions.JsonRpcInvalidJsonReceivedException(data)
        return data

    def _show_result(self, msg):
        pprint(msg)



#!/usr/bin/env python
#   BAREOS - Backup Archiving REcovery Open Sourced
#
#   Copyright (C) 2019-2020 Bareos GmbH & Co. KG
#
#   This program is Free Software; you can redistribute it and/or
#   modify it under the terms of version three of the GNU Affero General Public
#   License as published by the Free Software Foundation and included
#   in the file LICENSE.
#
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
#   Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301, USA.


from __future__ import print_function
import argparse
import bareos.bsock
import bareos.exceptions
import logging
import sys


def getArguments():
    argparser = argparse.ArgumentParser(description="Console to Bareos Director.")
    argparser.add_argument(
        "-d", "--debug", action="store_true", help="enable debugging output"
    )
    bareos.bsock.DirectorConsole.argparser_add_default_command_line_arguments(argparser)
    args = argparser.parse_args()
    return args


if __name__ == "__main__":
    logging.basicConfig(
        format="%(levelname)s %(module)s.%(funcName)s: %(message)s", level=logging.INFO
    )
    logger = logging.getLogger()

    args = getArguments()
    if args.debug:
        logger.setLevel(logging.DEBUG)

    bareos_args = bareos.bsock.DirectorConsole.argparser_get_bareos_parameter(args)
    logger.debug("options: %s" % (bareos_args))
    try:
        director = bareos.bsock.DirectorConsole(**bareos_args)
    except (bareos.exceptions.Error) as e:
        print(str(e))
        sys.exit(1)

    logger.debug("authentication successful")
    director.interactive()
